/* ======================================================================= */
/*  TEXAS INSTRUMENTS, INC.                                                */
/*                                                                         */
/*  FFTLIB  FFT Library                                                    */
/*                                                                         */
/* Copyright (C) 2013 Texas Instruments Incorporated - http://www.ti.com/  */ 
/*                                                                         */
/*                                                                         */
/*  Redistribution and use in source and binary forms, with or without     */
/*  modification, are permitted provided that the following conditions     */
/*  are met:                                                               */
/*                                                                         */
/*    Redistributions of source code must retain the above copyright       */
/*    notice, this list of conditions and the following disclaimer.        */
/*                                                                         */
/*    Redistributions in binary form must reproduce the above copyright    */
/*    notice, this list of conditions and the following disclaimer in the  */
/*    documentation and/or other materials provided with the               */
/*    distribution.                                                        */
/*                                                                         */
/*    Neither the name of Texas Instruments Incorporated nor the names of  */
/*    its contributors may be used to endorse or promote products derived  */
/*    from this software without specific prior written permission.        */
/*                                                                         */
/*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    */
/*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      */
/*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  */
/*  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   */
/*  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  */
/*  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       */
/*  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  */
/*  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  */
/*  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    */
/*  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  */
/*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   */
/*                                                                         */
/* ======================================================================= */

#include <ti/fftlib/src/common/fft_common.h>

#ifndef FFT_DP_1D_C2C_BATCH_H_
#define FFT_DP_1D_C2C_BATCH_H_ 1

/** @ingroup 1D */
/* @{ */

/** @defgroup FFT_DP_1D_C2C_BATCH */
/** @ingroup FFT_DP_1D_C2C_BATCH */
/* @{ */

/**
 *
 *     @param N = FFT size dimension 1        
 *     @param batch = FFT batch size
 *     @param mode = FFT_DIRECT & FFT_ECPY are supported for 1d FFT
 */

/* FFT plan 1D DP batch API */
fft_plan_t fft_dp_plan_1d_c2c_batch (
    int   N,        /* FFT size */
    int   batch,    /* batch size */
    int   mode,     /* FFT_DIRECT or FFT_ECPY */
    fft_callout_t fxns
);

 /**
 *     This program performs one 1-dimensional double-precision complex-to-complex batched FFT.
 *  
 *     @param N = FFT size      
 *     @param M = Power of 2 FFT size, if Bluestein algorithm is used      
 *     @param batch = FFT batch size
 *     @param rad = radix 2 or 4
 *     @param twsize = Twiddle buffer size in bytes
 *     @param N_p2 = Power of 2 FFT size in Mixed-radix FFT
 *     @param s_r3 = Number of stages of radix-3
 *     @param s_r5 = Number of states of radix-5
 *     @param in = input pointer    
 *     @param out = output pointer   
 *     @param tw = twiddle pointer  
 *     @param Bn = Bluestein pointer  
 *     @param workbuf = Bluestein scratch pointer
 * 
 * @par Algorithm:
 *      Algorithm supports radix values are 2, 3, 4, & 5, and Bluestein for all other sizes.<br>
 *
 * @par Assumptions:
 *      Batch size is at least 4.<br>
 *      N is a positive value.<br>
 *      All input, output and working buffers are double word aligned.<br>
 *      No input, working or buffers overlap.<br>
 *
 * @par Size Limitations:
 *      Current limitations are: None.
 *
 * @par Implementation notes:
 * @b Endian Support: The code supports little endian modes.  <BR> 
 * @b Interruptibility: The code is interrupt-tolerant but not interruptible. <BR>
 *  
 */

/* FFT execute 1D DP direct mode batch API */
int fft_dp_1d_c2c_batch_direct (
    /* int   N;        FFT size         */
    /* int   batch;    batch size       */
    /* int   rad;      radix            */
    /* double *in;      input pointer    */
    /* double *out;     output pointer   */
    /* double *tw;      twiddle pointer  */
    fft_param_u u,
    void  *edmaState
);

 /**
 *     This program performs one 1-dimensional double-precision complex-to-complex batched FFT.
 *  
 *     @param N = FFT size      
 *     @param M = Power of 2 FFT size, if Bluestein algorithm is used      
 *     @param batch = FFT batch size
 *     @param rad = radix 2 or 4
 *     @param twsize = Twiddle buffer size in bytes
 *     @param N_p2 = Power of 2 FFT size in Mixed-radix FFT
 *     @param s_r3 = Number of stages of radix-3
 *     @param s_r5 = Number of states of radix-5
 *     @param in = input pointer    
 *     @param in_w = working input pointer    
 *     @param out = output pointer   
 *     @param out_w = working output pointer    
 *     @param tw = twiddle pointer  
 *     @param tw_w = working twiddle pointer  
 *     @param Bn = Bluestein pointer  
 *     @param Bn_w = Bluestein buffer pointer
 *     @param workbuf = Bluestein scratch pointer
 * 
 * @par Algorithm:
 *      Algorithm supports radix values are 2, 3, 4, & 5, and Bluestein for all other sizes.<br>
 *
 * @par Assumptions:
 *      Batch size is at least 4.<br>
 *      N is a positive value.<br>
 *      All input, output and working buffers are double word aligned.<br>
 *      No input, working or buffers overlap.<br>
 *
 * @par Size Limitations:
 *      Current limitations are: None.
 *
 * @par Implementation notes:
 * @b Endian Support: The code supports little endian modes.  <BR> 
 * @b Interruptibility: The code is interrupt-tolerant but not interruptible. <BR>
 *  
 */

/* FFT execute 1D DP ecpy mode batch API */
int fft_dp_1d_c2c_batch_ecpy (
    /* int   N;        FFT size         */
    /* int   batch;    batch size       */
    /* int   rad;      radix            */
    /* double *in;      input pointer    */
    /* double *out;     output pointer   */
    /* double *in_w;    input pointer    */
    /* double *out_w;   output pointer   */
    /* double *tw;      twiddle pointer  */
    /* double *tw_w;    twiddle pointer  */
    fft_param_u u,
    void *edmaState
);
 
 /* }@ */ /* ingroup */
 /* }@ */ /* ingroup */

#endif /* FFT_DP_1D_C2C_BATCH_H_ */

/* ======================================================================== */
/*  End of file:  fft_dp_1d_c2c_batch.h                                     */
/* ------------------------------------------------------------------------ */
/*            Copyright (c) 2013 Texas Instruments, Incorporated.           */
/*                           All Rights Reserved.                           */
/* ======================================================================== */


