/* ======================================================================= */
/*  TEXAS INSTRUMENTS, INC.                                                */
/*                                                                         */
/*  FFTLIB  FFT Library                                                    */
/*                                                                         */
/* Copyright (C) 2013 Texas Instruments Incorporated - http://www.ti.com/  */ 
/*                                                                         */
/*                                                                         */
/*  Redistribution and use in source and binary forms, with or without     */
/*  modification, are permitted provided that the following conditions     */
/*  are met:                                                               */
/*                                                                         */
/*    Redistributions of source code must retain the above copyright       */
/*    notice, this list of conditions and the following disclaimer.        */
/*                                                                         */
/*    Redistributions in binary form must reproduce the above copyright    */
/*    notice, this list of conditions and the following disclaimer in the  */
/*    documentation and/or other materials provided with the               */
/*    distribution.                                                        */
/*                                                                         */
/*    Neither the name of Texas Instruments Incorporated nor the names of  */
/*    its contributors may be used to endorse or promote products derived  */
/*    from this software without specific prior written permission.        */
/*                                                                         */
/*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    */
/*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      */
/*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  */
/*  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   */
/*  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  */
/*  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       */
/*  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  */
/*  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  */
/*  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    */
/*  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  */
/*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   */
/*                                                                         */
/* ======================================================================= */

#include <xdc/std.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <c6x.h>

#include <ti/fftlib/src/common/omp/omp_config.h>
#include "ifft_omp_dp_1d_c2c.h"
#include <ti/fftlib/src/common/fft_edma.h>
#include <ti/dsplib/src/DSPF_dp_ifftDPxDP/DSPF_dp_ifftDPxDP.h>

#include <ti/runtime/openmp/omp.h>

extern cregister volatile unsigned int DNUM;
#define ACTIVE_THREAD_COUNT (OMP_NUM_CORES)

#define FFT_EDMA_STATE_INIT         0
#define FFT_EDMA_STATE_ALLOCATED    1

int fftEdmaState[OMP_MAX_NUM_CORES];
FFT_EDMA_Struct gEdmaState[OMP_MAX_NUM_CORES];
#pragma DATA_SECTION (gEdmaState, ".ddr_heap")
/* Function to trap exceptions */
void fft_assert(int statement, int node_id, const char *error)
{
  volatile int dbg_halt = 1;

  if(!statement) {
    printf("%s (%d)\n",error,node_id);
    while(dbg_halt);
  }
}

void *fft_omp_assign_edma_resources(void)
{
  /*
   * The edmaInstances are indexes into the C6678_config[] array defined in
   * fft_c6678_config, which is used to specify how EDMA resources are
   * divided between cores.
   */
  int   ret_val;
  void *ret = (void *) (&gEdmaState[0]);

  #pragma omp parallel
  {
    if ( fftEdmaState[DNUM] != FFT_EDMA_STATE_ALLOCATED )
    {
      gEdmaState[DNUM].num_channels = 0;

      while ( gEdmaState[DNUM].num_channels < FFT_NUM_EDMA_CH )
      {
        fft_assert( ((gEdmaState[DNUM].channel[gEdmaState[DNUM].num_channels]) = EdmaMgr_alloc(FFT_MAX_EDMA_LINKS)) != NULL , DNUM, "EdmaMgr_alloc() failed ");

        gEdmaState[DNUM].num_channels++;
      }
    }
    fftEdmaState[DNUM] = FFT_EDMA_STATE_ALLOCATED;
  }

  return ret;
}

void fft_omp_free_edma_resources(void *edma)
{
  /*
   * The edmaInstances are indexes into the C6678_config[] array defined in
   * fft_c6678_config, which is used to specify how EDMA resources are
   * divided between cores.
   */
  int ret_val;

  #pragma omp parallel
  {

    if ( fftEdmaState[DNUM] == FFT_EDMA_STATE_ALLOCATED )
    {
      while ( gEdmaState[DNUM].num_channels > 0 )
      {
        gEdmaState[DNUM].num_channels--;

        ret_val = EdmaMgr_free(gEdmaState[DNUM].channel[gEdmaState[DNUM].num_channels]);

        fft_assert( ret_val == EdmaMgr_SUCCESS, DNUM, "EDMA free failed!");

      }
    }
    fftEdmaState[DNUM] = FFT_EDMA_STATE_INIT;
  }
}

void tw_gen_cn (double *w, int n);
void idft (int N, double x[], double y[], int N1);

/* ======================================================================== */
/*  Kernel-specific alignments                                              */
/* ======================================================================== */
#pragma DATA_SECTION(x_i, ".mem_ddr");
#pragma DATA_SECTION(y_i, ".mem_ddr");
#pragma DATA_SECTION(w_i, ".mem_ddr");
#pragma DATA_SECTION(x_cn, ".mem_ddr");
#pragma DATA_SECTION(y_cn, ".mem_ddr");
#pragma DATA_SECTION(w_cn, ".mem_ddr");

#pragma DATA_ALIGN(x_i,  8);
#pragma DATA_ALIGN(x_cn, 8);

#pragma DATA_ALIGN(w_cn, 8);

#pragma DATA_ALIGN(y_i,  8);
#pragma DATA_ALIGN(y_cn, 8);

#pragma DATA_SECTION(x_i_work, ".mem_l2");
#pragma DATA_SECTION(y_i_work, ".mem_l2");
#pragma DATA_SECTION(y_i_temp, ".mem_l2");
#pragma DATA_SECTION(w_i_work, ".mem_l2");

#pragma DATA_ALIGN(w_i,  8);
#pragma DATA_ALIGN(w_i_work,  8);
#pragma DATA_ALIGN(x_i_work,  64);
#pragma DATA_ALIGN(y_i_work,  64);

/* ======================================================================== */
/*  Parameters of fixed dataset.                                            */
/* ======================================================================== */

#define MAXN  (1024*1024)
#define M     (2*MAXN)
#define M_i   (4*1024)
#define PAD   (0)

/* ======================================================================== */
/*  Initialized arrays with fixed test data.                                */
/* ======================================================================== */

double x_i [M + 2 * PAD];
double x_cn[M + 2 * PAD];

double y_i [M + 2 * PAD];
double y_cn[M + 2 * PAD];

double x_i_work [M_i*4 + 2 * PAD];
double y_i_work [M_i*4 + 2 * PAD];
double y_i_temp [M_i*2 + 2 * PAD];
double w_i_work [1024*6 + 2 * PAD];

double w_i [1024*6 + 2 * PAD];
double w_cn[M + 2 * PAD];

/* ======================================================================== */
/*  Generate pointers to skip beyond array padding                          */
/* ======================================================================== */
double *const ptr_x_i  = x_i  + PAD;
double *const ptr_x_cn = x_cn + PAD;

double *const ptr_w_i  = w_i  + PAD;
double *const ptr_w_cn = w_cn + PAD;

double *const ptr_y_i  = y_i  + PAD;
double *const ptr_y_cn = y_cn + PAD;

double *const ptr_y_i_temp = y_i_temp + PAD;
double *const ptr_y_i_work  = y_i_work + PAD;
double *const ptr_x_i_work  = x_i_work + PAD;
double *const ptr_w_i_work  = w_i_work + PAD;

void fft_memory_request (int nbufs, FFTmemBuffer_t *bufs)
{
  int i;

  printf ("FFT memory buffers:\n");
  printf ("    Buffer    Size(bytes)    Alignment\n");
  for (i = 0; i < nbufs; i++) {
    printf ("     %3d       %7d         %4d       \n", i, (int)bufs[i].size, (int)bufs[i].log2align);
  }
  bufs[0].base = ptr_x_i;
  bufs[1].base = ptr_y_i;
  bufs[2].base = ptr_w_i;

  bufs[3].base = ptr_x_i_work;
  bufs[4].base = ptr_y_i_work;
  bufs[5].base = ptr_w_i_work;
  bufs[6].base = ptr_y_i_temp;

} /* fft_memory_request */

void fft_memory_release (int nbufs, FFTmemBuffer_t *bufs)
{
  /* do nothing for now */
} /* fft_memory_request */

/* ======================================================================== */
/*  MAIN -- Top level driver for the test.                                  */
/* ======================================================================== */

int main ()
{
    int     i, j, N, rad_cn;

    /* BaseN calculation based upon number of cores & cache lines for 4step fft */
    int     baseN = ACTIVE_THREAD_COUNT*ACTIVE_THREAD_COUNT*IFFT_OMP_DP_1D_C2C_NUMOFLINEBUFS*IFFT_OMP_DP_1D_C2C_NUMOFLINEBUFS;
    int     rad3 = 1, rad5 = 1, rad15 = 1;
    clock_t t_start, t_stop, t_overhead, t_opt;
    double  diff, max_diff = 0;
    fft_plan_t p;
    fft_callout_t plan_fxns;

    /* --------------------------------------------------------------------- */
    /* intialize hardware timers                                             */
    /* --------------------------------------------------------------------- */
    TSCL=0;TSCH=0;

    /* initalize callout functions */
    plan_fxns.memoryRequest   = fft_memory_request;
    plan_fxns.memoryRelease   = fft_memory_release;
    plan_fxns.ecpyRequest     = fft_omp_assign_edma_resources;
    plan_fxns.ecpyRelease     = fft_omp_free_edma_resources;
    
    /* initialize ECPY */
    omp_set_num_threads (OMP_MAX_NUM_CORES);

    #pragma omp parallel
    {
      fft_assert( (EdmaMgr_init(DNUM, NULL) == EdmaMgr_SUCCESS), DNUM, "EdmaMgr_init() return error!");
      fftEdmaState[DNUM] = FFT_EDMA_STATE_INIT;
    }

    /* radix 2&4 testing */
    for (N = 1024; N <= MAXN; N = N*2)
    {

        memset (x_i,  0x55, sizeof (x_i) );
        memset (x_cn, 0x55, sizeof (x_cn));

        /* ---------------------------------------------------------------- */
        /* Initialize input vector temporarily.                             */
        /* ---------------------------------------------------------------- */

        for (i = 0; i < N; i++) {
          x_cn[PAD + 2*i  ] = sin (2 * 3.1415 * 50 * i / (double) N);
          x_cn[PAD + 2*i+1] = sin (2 * 3.1415 * 60 * i / (double) N);
        }
        for (j = 0; j < 2*N; j++) {
          x_i[PAD + j] = x_cn[PAD + j];
        }
        
        /* ---------------------------------------------------------------- */
        /* Force uninitialized arrays to fixed values.                      */
        /* ---------------------------------------------------------------- */
        memset (y_i,  0xA5, sizeof (y_i) );
        memset (y_cn, 0xA5, sizeof (y_cn));

        /* ---------------------------------------------------------------- */
        /* Generate twiddle factors.                                        */
        /* ---------------------------------------------------------------- */
        j = 0;
        for (i = 0; i <= 31; i++)
            if ((N & (1 << i)) == 0)
                j++;
            else
                break;

        if (j % 2 == 0) {
            rad_cn = 4;
        }
        else {
            rad_cn = 2;
        }

        tw_gen_cn (ptr_w_cn, N);
        t_start = _itoll(TSCH, TSCL);
        DSPF_dp_ifftDPxDP (N, ptr_x_cn, ptr_w_cn, ptr_y_cn, rad_cn, 0, N);
        t_stop = _itoll(TSCH, TSCL);
        //        dft_dp (N, ptr_x_cn, ptr_y_cn, N);
        p = ifft_omp_dp_plan_1d_c2c (N, FFT_ECPY, plan_fxns);
        
        /* ---------------------------------------------------------------- */
        /* Compute the overhead of allocating and freeing EDMA              */
        /* ---------------------------------------------------------------- */
        p.edmaState = (*p.fftcout.ecpyRequest)();
        (*p.fftcout.ecpyRelease)(p.edmaState);
        p.edmaState = (*p.fftcout.ecpyRequest)();
        (*p.fftcout.ecpyRelease)(p.edmaState);
        t_start = _itoll(TSCH, TSCL);
        p.edmaState = (*p.fftcout.ecpyRequest)();
        (*p.fftcout.ecpyRelease)(p.edmaState);
        t_stop  = _itoll(TSCH, TSCL);
        t_overhead = t_stop - t_start;

        /* ---------------------------------------------------------------------- */
        /* Set the number of cores used                                           */
        /* ---------------------------------------------------------------------- */
        p.u.dp_1d_c2c_e.actualCoreNum = OMP_MAX_NUM_CORES;

        /***************************************
         * ecpy fft test
         ***************************************/
        t_start = _itoll(TSCH, TSCL);
        fft_execute (p);
        t_stop = _itoll(TSCH, TSCL);

        fft_destroy_plan (p);
        t_opt  = (t_stop - t_start) - t_overhead;

        /* ---------------------------------------------------------------- */
        /* compute difference and track max difference                      */  
        /* ---------------------------------------------------------------- */
        diff = 0; max_diff = 0;
        for(i=0; i<2*N; i++) {
          diff = _fabs(ptr_y_cn[i] - ptr_x_i[i]);
          if (diff > max_diff) max_diff = diff;
        }
        printf("ifft_omp_dp_1d_c2c_ecpy\tsize = %d\n", N);
        printf("max_diff = %f", max_diff);
        printf("\tN = %d\tCycle: %d\n\n", N, t_opt);
    }
    
/* Setup initial starting points for MC ECPY */
while (rad3*baseN  < IFFT_OMP_DP_1D_C2C_4STEP_MIN_SIZE) rad3*=3;
while (rad5*baseN  < IFFT_OMP_DP_1D_C2C_4STEP_MIN_SIZE) rad5*=5;
while (rad15*baseN < IFFT_OMP_DP_1D_C2C_4STEP_MIN_SIZE) rad15*=15;
    /* radix 3 & 5 testing */
    while ( rad15*baseN <= 10000)
    {
    	/*generate a sequence of numbers guaranteed to be radix 3, 5 or 3 & 5 */
    	if ( rad3*baseN < 10000){
    		N = baseN*rad3;
    		rad3*=3;
    	} else if ( rad5*baseN < 10000) {
    		N = baseN*rad5;
    		rad5*=5;
    	} else {
    		N = baseN*rad15;
    		rad15*=15;
    	}

        memset (x_i,  0x55, sizeof (x_i) );
        memset (x_cn, 0x55, sizeof (x_cn));

        /* ---------------------------------------------------------------- */
        /* Initialize input vector temporarily.                             */
        /* ---------------------------------------------------------------- */

        for (i = 0; i < N; i++) {
          x_cn[PAD + 2*i  ] = sin (2 * 3.1415 * 50 * i / (double) N);
          x_cn[PAD + 2*i+1] = sin (2 * 3.1415 * 60 * i / (double) N);
        }
        for (j = 0; j < 2*N; j++) {
          x_i[PAD + j] = x_cn[PAD + j];
        }
        
        /* ---------------------------------------------------------------- */
        /* Force uninitialized arrays to fixed values.                      */
        /* ---------------------------------------------------------------- */
        memset (y_i,  0xA5, sizeof (y_i) );
        memset (y_cn, 0xA5, sizeof (y_cn));
        
        idft (N, ptr_x_cn, ptr_y_cn, N);
        p = ifft_omp_dp_plan_1d_c2c (N, FFT_ECPY, plan_fxns);
        
        /* ---------------------------------------------------------------- */
        /* Compute the overhead of allocating and freeing EDMA              */
        /* ---------------------------------------------------------------- */
        p.edmaState = (*p.fftcout.ecpyRequest)();
        (*p.fftcout.ecpyRelease)(p.edmaState);
        p.edmaState = (*p.fftcout.ecpyRequest)();
        (*p.fftcout.ecpyRelease)(p.edmaState);
        t_start = _itoll(TSCH, TSCL);
        p.edmaState = (*p.fftcout.ecpyRequest)();
        (*p.fftcout.ecpyRelease)(p.edmaState);
        t_stop  = _itoll(TSCH, TSCL);
        t_overhead = t_stop - t_start;

        /* ---------------------------------------------------------------------- */
        /* Set the number of cores used                                           */
        /* ---------------------------------------------------------------------- */
        p.u.dp_1d_c2c_e.actualCoreNum = OMP_MAX_NUM_CORES;

        /***************************************
         * ecpy fft test
         ***************************************/
        t_start = _itoll(TSCH, TSCL);
        fft_execute (p);
        t_stop = _itoll(TSCH, TSCL);

        fft_destroy_plan (p);
        t_opt  = (t_stop - t_start) - t_overhead;

        /* ---------------------------------------------------------------- */
        /* compute difference and track max difference                      */  
        /* ---------------------------------------------------------------- */
        diff = 0; max_diff = 0;
        for(i=0; i<2*N; i++) {
          diff = _fabs(ptr_y_cn[i] - ptr_x_i[i]);
          if (diff > max_diff) max_diff = diff;
        }
        printf("ifft_omp_dp_1d_c2c_ecpy\tsize = %d\n", N);
        printf("max_diff = %f", max_diff);
        printf("\tN = %d\tCycle: %d\n\n", N, t_opt);
    }    
}

/* Function for generating Specialized sequence of twiddle factors */
void tw_gen_cn (double *w, int n)
{
    int i, j, k;
    const double PI = 3.141592654;

    for (j = 1, k = 0; j <= n >> 2; j = j << 2)
    {
        for (i = 0; i < n >> 2; i += j)
        {
            w[k]     = cos (2 * PI * i / n);
            w[k + 1] = sin (2 * PI * i / n);
            w[k + 2] = cos (4 * PI * i / n);
            w[k + 3] = sin (4 * PI * i / n);
            w[k + 4] = cos (6 * PI * i / n);
            w[k + 5] = sin (6 * PI * i / n);
            k += 6;
        }
    }
}

/* Function for calculating any size iDFT */
void idft (int N, double x[], double y[], int N1)
{
    int k, i, index;
    const double PI = 3.141592654;
    double *p_x;
    double arg, fx_0, fx_1, fy_0, fy_1, co, si;

    for (k = 0; k < N1; k++)

    {
        p_x = x;

        fy_0 = 0;
        fy_1 = 0;
        for (i = 0; i < N; i++)
        {
            fx_0 = p_x[0];
            fx_1 = p_x[1];
            p_x += 2;
            index = (i * k) % N;
            arg = 2 * PI * index / N;
            co = cos (arg);
            si = sin (arg);
            fy_0 += ((fx_0 * co) - (fx_1 * si));
            fy_1 += ((fx_1 * co) + (fx_0 * si));
        }
        y[2 * k] = fy_0/N;
        y[2 * k + 1] = fy_1/N;
    }
}

/* ======================================================================== */
/*  End of file:  ifft_omp_dp_1d_c2c_d.c                                    */
/* ------------------------------------------------------------------------ */
/*            Copyright (c) 2013 Texas Instruments, Incorporated.           */
/*                           All Rights Reserved.                           */
/* ======================================================================== */
